/*!
 * (C) Copyright IBM Corp. 2005, 2005
 *
 * \file kboot.c
 * \brief main function for kboot boot loader
 *
 * $Id: kboot.c,v 1.16 2006/06/01 13:56:08 loehr Exp $
 *
 * Author(s): Ralph Wuerthner (rwuerthn@de.ibm.com)
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>       //basename
#include <syslog.h>       //syslog
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "kboot.h"
#include "loader.h" // only for debugging - remove later!


char *arg0; //<! global variable pointing to argv[0] (used in MEM_ASSERT)

/**
 * Print help message
 *
 * \param[in] arg0 \p argv[0] from \p main()
 */

void
print_help(const char *arg0)
{
	printf("Usage: %s [OPTION] <configuration file URI>\n"
	    "Start kboot user interface.\n\n"
	    " -o <device>   Output kboot status messages to <device>. "
	    "Default is stdout.\n"
	    " -u <uitype>   Ignore global definitions/Start only this ui. \n"
	    " -v            Output version information and exit.\n"
	    " -h            Display this help and exit.\n",
	    arg0);
}


/**
 * Redirect stdout & stderr to different device if requested on command line.
 *
 * \param[in]  device Path to output device.
 */

void
redirect_output(const char *device)
{
	int fd;
	struct stat st;

	if (strlen(device)) {
		if (stat(device, &st)) {
			printf("Warning: unable to check output device '%s'"
			    " - %s\n", device, strerror(errno));
			goto error;
		}
		if ((st.st_mode & S_IFMT) != S_IFCHR) {
			printf("Warning: '%s' is not a character device\n",
			    device);
			goto error;
		}
		fd = open(device, O_RDWR | O_NONBLOCK, 0);
		if (fd != -1) {
			dup2(fd, 1);
			dup2(fd, 2);
			close(fd);
			return;
		}
		printf("Warning: error opening output device '%s' - %s\n",
		    device, strerror(errno));

	error:
		printf("Continuing using stdout as output device.\n");
	}
}


/**
 * \p main -- the entry point into a world of wonderful possibilities.
 *
 * \param[in]  argc Number of arguments passed to program.
 * \param[in]  argv Array with arguments passed to program.
 */

int
main(int argc, char **argv)
{
	char *errmsg, *startup_msg;
	struct kboot_arguments kboot_args;
	struct cfg_toplevel config;
	struct cfg_bentry boot;
	enum parse_result pa_return;
	int retval = 0; 

	arg0 = argv[0];
        dg_level = DG_MINIMAL;

        openlog(basename(argv[0]), LOG_PID | LOG_CONS, LOG_USER);

        dg_init();

	pa_return = parse_arguments(&kboot_args, argc, argv);
	dg_printf(DG_VERBOSE, "%s:parse_arguments done.\n", __FUNCTION__);

	switch (pa_return) {

	case PA_DONE:
		retval = 0;
		break;

	case PA_ERROR:
		retval = 1;
		break;

	case PA_VALID:
		redirect_output(kboot_args.console);
		dg_printf(DG_VERBOSE, 
		    "%s:redirect output done.\n", __FUNCTION__);
		cfg_strinit(&startup_msg);

		// print startup message to redirected console
		syslog(LOG_INFO,"kboot user interface is starting.");
		syslog(LOG_INFO,"Configuration file source: %s",
		    kboot_args.config_uri);

		// handle kboot config info on the kernel command line
		if (strlen(kboot_args.only_ui) == 0)
			parse_kernel_cmdline();

		// parse configuration file
		errmsg = parse_kboot(&config, &kboot_args);
		if (errmsg) {
			cfg_strprintf(&startup_msg,
			    "Error loading configuration data:\n"
			    "%s\n"
			    "Continuing with blank configuration\n", errmsg);
			free(errmsg);
			cfg_init(&config);
		}

		while (WORLD_EXISTS) {
			// launch user interface modules
			if (userinterface(startup_msg, &config, &boot)) {
				syslog(LOG_ERR,
				    "unable to find boot configuration\n"
				    "kboot is exiting - goodbye!");
				retval = 1;
				break;
			}
			cfg_strcpy(&startup_msg, "");

			// start new kernel
			errmsg = loader(&boot);
			if (errmsg) {
				cfg_strprintf(&startup_msg,
				    "Unable to  start selected configuration:\n"
				    "%s\n"
				    "Please try again.\n",
				    errmsg);
				free(errmsg);
			} else
				cfg_strprintf(&startup_msg,
				    "Error starting selected configuration. "
				    "Please try again.\n");
			cfg_bentry_destroy(&boot);
		}
		break;
	}

	// enter a shell for debugging purposes
	action_shell();
        closelog();
	// always reached

	return retval;
}
