#!/usr/bin/perl
# ---------------------------------------------------------------------------
# (C) Copyright IBM Corporation 2006, 2006
# 
# \file kboot_admin.pl.
#
# Author(s): Swen Schillig (swen@vnet.ibm.com)
# ---------------------------------------------------------------------------
# $Id: kboot_admin.pl,v 1.18 2006/06/09 09:00:34 swen Exp $
# ---------------------------------------------------------------------------

use strict;
use Getopt::Long;
use File::Copy;
use File::Basename;
use File::Spec::Functions;
use File::Path;
use Cwd;

our $VERSION = "0.1";

############### command line options/switches ###############################
my $conf_file;
my $master;
my $kboot;
my $image;
my $kernel;
my @modules;
my @executables;
my @libraries;
my $help;
my $vers;
#############################################################################

my $GZIP;
my $FIND;
my $CPIO;
my $CUT;
my $RSYNC;
my $GK_VERS;
my $t_dir;

Getopt::Long::Configure("bundling");
exit(1) unless( GetOptions('c|config=s' => \$conf_file, 'm|master=s' => \$master,
                           's|secondary=s' => \$kboot, 'o|output=s' => \$image,
                           'l|lib=s' => \@libraries, 'd|module=s' => \@modules,
                           'k|kernel=s' => \$kernel, 'x|executable=s' => \@executables,
                           'h|help' => \$help, 'v|version' => \$vers)
              );

$conf_file = "/etc/kboot_admin.conf" unless $conf_file;

unless($< == 0)  # are we super user ?
{
    print("You have to be super user to execute ",basename($0),"\n");
    exit(-1);
}

if($vers)
{
    print basename($0)," Version: $VERSION\n";
    exit(0);
}

if(! @ARGV or $help)
{
    #print "Please specify a command <merge|local|passwd|add>\n" unless $help;
    print "Please specify a command <merge|passwd|add>\n" unless $help;
    usage();
    exit(0);
}

@executables = split(/[,\s]/,join(',',@executables));
@libraries   = split(/[,\s]/,join(',',@libraries));
@modules     = split(/[,\s]/,join(',',@modules));

init() or do {
    print("Please verify your configuration file $conf_file and\n");
    print("add the mandatory entries manually.\n");
    exit(1);
};

if("merge local passwd add" !~ /\b$ARGV[0]\b/)
{
    print("Unknown command '$ARGV[0]'\n\n");
    usage();
    exit(50);
}

if( ("merge add" =~ /$ARGV[0]/) && (! -e $kernel))
{
    print "Kernel '$kernel' does not exist.\n";
    exit(1);
}

print("Unpacking master ramfs '$master'\n");
unpack_ramfs($master,catdir($t_dir,"master")) or exit(2);

if($ARGV[0] eq "passwd")
{
    print("Changing password\n");
    change_password(catdir($t_dir,"master"))  or exit(3);
}
elsif($ARGV[0] eq "merge")
{
    print("Unpacking secondary ramfs '$kboot'\n");
    unpack_ramfs($kboot,  catdir($t_dir,"kboot")) or exit(6);
    
    print("Synchronizing ramfs'\n");
    merge_ramfs(catdir($t_dir,"master"), catdir($t_dir,"kboot")) or exit(7);
}
elsif($ARGV[0] eq "local")
{
    print "\nUNDER CONSTRUCTION !!\n";
}
elsif($ARGV[0] eq "add")
{
    if(@modules)
    {
        print("Adding modules to ramfs\n");
        add_modules(catdir($t_dir,"master"), $kernel, @modules) or do {
            print("Cannot add modules to ramfs\n");
            exit(6);
        };
    }
    
    if(@executables)
    {
        print("Adding executables to ramfs\n");
        add_executables(catdir($t_dir,"master"), @executables) or do {
            print "Cannot add executables to ramfs\n";
            exit(6);
        };
    }
    
    if(@libraries)
    {
        print("Adding libraries to ramfs\n");
        add_libraries(catdir($t_dir,"master"), @libraries) or do {
            print "Cannot add libraries to ramfs\n";
            exit(6);
        };
    }
}

print("Creating ramfs '$image'\n");
pack_ramfs($image, catdir($t_dir,"master"))   or exit(4);

print("Cleaning up ...");
rmtree($t_dir, 0, 0);
print("\n\n");


################## SUB's ############################################
sub usage
{
    $0 = basename($0);
    #print "\n[SYNTAX]  $0 [Options] <merge|local|passwd|add>\n\n";
    print "\n[SYNTAX]  $0 [Options] <merge|passwd|add>\n\n";
    print "$0 is used to modify the initial ramfs by \n";
    print " - adding modules, libraries or executables,\n";
    print " - merging two ramfs' (e.g. a distro-ramfs and a kboot-ramfs)\n";
    print " - changing the password for the super-user (for ssh login).\n";
    #print "The command 'local' is primarily used to add the extended boot functionality\n";
    #print "to an existing distribution ramfs using the locally available files\n\n";
    print "Commands:\n";
    print "  merge    join content from -s <ramfs> and -m <ramfs> to -o <ramfs>\n";
    #print "  local    create a -o <ramfs> based on the running systems and the -m <ramfs>\n";
    print "  passwd   change password for UID kboot in -m <ramfs>\n";
    print "  add      add executables, libraries and/or modules to the ramfs\n";
    print "Options:\n";
    print "  -h|--help                this help text\n";
    print "  -v|--version             print version information\n";
    print "  -c|--config     <file>   configuration file name (default /etc/kboot_admin.conf)\n";
    print "  -m|--master     <file>   master ramfs superseeding other ramfs' content.\n";
    print "  -s|--secondary  <file>   second ramfs which is merged with the master ramfs.\n";
    print "  -o|--output     <file>   file name for the resulting ramfs image\n";
    print "  -k|--kernel     <file>   kernel file name\n";
    print "  -l|--lib        <file>   add DLL(s) <file,...> to ramfs (full qualified path)\n";
    print "  -m|--module     <module> add module(s) <module,...> to ramfs (e.g. qeth )\n";
    print "  -x|--executable <file>   add executable <file,...> to ramfs (e.g. losetup)\n";
}

sub init
{                #do we have cmd_line options
    my $extern = ! (@executables or @libraries or @modules);
    
    unless( -r $conf_file && open(FH, "<", $conf_file))
    {
        print("Cannot read configuration file '$conf_file'\n");
        print("Do you want to generate a new config file [No/yes]? ");
        
        return undef if(getc() !~ /[yY]/);
        
        return undef unless initialize($conf_file);
        open(FH, "<", $conf_file);
    }
    
    while(<FH>)
    {            
        next if(/^\s*(#.*)?$/); #goto next line if only spaces/comments
        s/^\s*(.*)\s*#.*$/$1/;  #remove leading spaces and trailing comments/spaces
        chomp;
        
        /^GZIP/         && do { ($GZIP     = $_) =~ s/GZIP[\s=]*(.*)/$1/;       next; };
        /^FIND/         && do { ($FIND     = $_) =~ s/FIND[\s=]*(.*)/$1/;       next; };
        /^CPIO/         && do { ($CPIO     = $_) =~ s/CPIO[\s=]*(.*)/$1/;       next; };
        /^CUT/          && do { ($CUT      = $_) =~ s/CUT[\s=]*(.*)/$1/;        next; };
        /^RSYNC/        && do { ($RSYNC    = $_) =~ s/RSYNC[\s=]*(.*)/$1/;      next; };
        /^GET_K_VERS/   && do { ($GK_VERS  = $_) =~ s/GET_K_VERS[\s=]*(.*)/$1/; next; };
        /^TEMP_DIR/     && do { ($t_dir    = $_) =~ s/TEMP_DIR[\s=]*(.*)/$1/;   next; };
        
        if($extern) #no cmd_line params specified,
        {           #so read conf_file additions
            /^MODULE[\s=]+/ && do { s/MODULE[\s=]+(.*)/$1/; push @modules, $_;   next; };
            /^EXEC[\s=]+/   && do { s/EXEC[\s=]+(.*)/$1/; push @executables, $_; next; };
            /^LIB[\s=]+/    && do { s/LIB[\s=]+(.*)/$1/; push @libraries, $_;    next; };
        }
        /^KERNEL/       && do { ($kernel = $_) =~ s/KERNEL[\s=]*(.*)/$1/       unless $kernel; next; };
        /^MASTER_RAMFS/ && do { ($master = $_) =~ s/MASTER_RAMFS[\s=]*(.*)/$1/ unless $master; next; };
        /^KBOOT_RAMFS/  && do { ($kboot  = $_) =~ s/KBOOT_RAMFS[\s=]*(.*)/$1/  unless $kboot;  next; };
        /^IMAGE_RAMFS/  && do { ($image  = $_) =~ s/IMAGE_RAMFS[\s=]*(.*)/$1/  unless $image;  next; };
    }
    
    -x $GZIP    || do { print "GZIP('$GZIP') is not executable.\n";        return undef; };
    -x $FIND    || do { print "FIND('$FIND') is not executable.\n";        return undef; };
    -x $CPIO    || do { print "CPIO('$CPIO') is not executable.\n";        return undef; };
    -x $CUT     || do { print "CUT('$CUT') is not executable.\n";          return undef; };
    -x $RSYNC   || do { print "RSYNC('$RSYNC') is not executable.\n";      return undef; };
    -x $GK_VERS || do { print "GET_K_VERS('$GK_VERS') is not executable.\n";
                        print "Please make sure an executable similar to 'get_kernel_version'\n";
                        print "is made available (check the internet for sources).\n";
                        return undef; };
    
    return close(FH);
}

sub initialize
{
    my $cf = shift;
    my $temp;
    my $error = 0;
    
    open(FH, ">", $cf) or do { print "Cannot create config-file '$cf': $!\n"; return undef;};
    
    chomp($temp = `which gzip`);
    print FH "GZIP       $temp\n";
    $error++ unless -x $temp;
        
    chomp($temp = `which find`);
    print FH "FIND       $temp\n";
    $error++ unless -x $temp;
    
    chomp($temp = `which cpio`);
    print FH "CPIO       $temp\n";
    $error++ unless -x $temp;
    
    chomp($temp = `which cut`);
    print FH "CUT        $temp\n";
    $error++ unless -x $temp;
    
    chomp($temp = `which rsync`);
    print FH "RSYNC      $temp\n";
    $error++ unless -x $temp;
    
    chomp($temp = `which get_kernel_version`);
    print FH "GET_K_VERS $temp\n";
    $error++ unless -x $temp;
    
    while(<DATA>)
    {
        print FH $_;
    }
    close FH;
    print("The automatic configuration finished.\n");
    print("Please verify the settings in the '$cf' file before you proceed.\n");
          
    if($error)
    {
        print "At least one mandatory command cannot be found.\n",
        "Please update the config-file '$cf' manually.\n";
        return undef;
    }
    return 1;
}

sub change_password
{
    my $directory = shift;
    
    system("cd $directory && chroot . /usr/bin/passwd") == 0 or do
    {
        print("Cannot change passwd.\n");
        return undef;
    };
    
    return 1;
}

sub unpack_ramfs
{
    my $image = shift;
    my $dest  = shift;
    my $temp  = getcwd();
    
    if( -e $dest)
    {
        if( ! -d $dest )
        {
            print("Destination exists and is not a directory.\n");
            return undef;
        }
        rmtree($dest);
        if( -d $dest)
        {
            print("Cannot delete directory content.\n");
            return undef;
        }
    }
    eval { mkpath($dest) };
    if($@) 
    {
        print "Cannot create $dest: $@\n";
        return undef;
    }
    
    if(! -r $image)
    {
        print("Cannot read image '$image'.\n");
        return undef;
    };
    
    copy($image,$dest) or do {
        print("Cannot copy image file to '$dest'.\n");
        return undef;
    };
    
    system("cd $dest && $GZIP -dc ".catfile($dest,basename($image))."|$CPIO -idu >/dev/null 2>&1") == 0 or 
        do {
            print "Cannot unpack ramfs: $!\n";
            return undef;
        };
    
    return unlink catfile($dest,basename($image));
}

sub pack_ramfs
{
    my $image  = shift;
    my $source = shift;
    
    if( ! -d $source )
    {
        print("'$source' is not a directory.\n");
        return undef;
    }
    
    $image = catfile(getcwd(),$image) unless $image =~ /^[\/~]/ ;
        
    if(-e $image and ! -w $image)
    {
        print("Cannot write ramfs image '$image'.\n");
        return undef;
    }
    
    system("cd $source && $FIND . -print | $CPIO -oc 2>/dev/null |$GZIP --best >$image") == 0 or 
        do {
            print "Cannot build ramfs: $!\n";
            return undef;  
        };
        
    return 1;
}

sub merge_ramfs
{
    my $master = shift;
    my $source = shift;
    
    unlink(catfile($source,"init"));
    
    system("cd $master && $FIND . -not -type d| $CUT -c 3- >../exclude_list") == 0 or do
    {
        print("Cannot generate exclude list for merging both ramfs'\n");
        return undef;
    };
    
    system("cd $t_dir && $RSYNC -a --ignore-existing --exclude-from=exclude_list $source $master") == 0 or do
    {
        print("Cannot merge ramfs: $!\n");
        return undef;
    };
    
    eval { symlink("/sbin/init.kboot", catfile($master,"vendor_init.sh")); 1 } or do
    {
        print("Cannot create symbolic link.\n");
        return undef;
    };
}

sub add_modules
{
    my $master = shift;
    my $kernel = shift;
    my @mod_list;
    my @mod_dep;
    my %check = {};
    my @mod    = map { /([^\s]*).*/ } @_;
    my $k_version = `$GK_VERS $kernel` or return undef;
    chomp($k_version);
    
    open(FH,"<","/lib/modules/$k_version/modules.dep") or 
        do {
            print "Cannot open file '/lib/modules/$k_version/modules.dep': $!\n";
            return undef;
        };
    @mod_dep = <FH>;
    close FH;

    eval { mkpath(dirname(catfile($master, "/lib/modules/$k_version/modules.dep"))) };
    if($@)
    {
        print "Cannot create directory: $@\n";
        return undef;
    }
    
    open(FH2,">",catfile($master, "/lib/modules/$k_version/modules.dep")) or die "EEEEE $!\n";
    eval { mkpath(dirname(catfile($master, "/lib/modules/$k_version/modules.dep"))) };
    if($@)
    {
        print "Cannot create directory: $@\n";
        return undef;
    }

    
    open(FH2,">",catfile($master, "/lib/modules/$k_version/modules.dep"));
    
    foreach my $module (@mod)
    {
        foreach my $line (@mod_dep)
        {
            if($line =~ /$module.ko:/)
            {
                print FH2 $line;  #add entry to new modules.conf
                $check{$module} = 1;
                $line =~ s/://;
                my @dummy = split(/\s+/,$line);
                push @mod_list, @dummy;
                map { s/.*\/([\w-]*)\.ko.*/$1/ } @dummy;
                push @mod, @dummy[1..$#dummy];
            }
        }
    }
    
    close FH2;
    
    foreach my $module (@mod_list)
    {
        eval { mkpath(catdir($master,dirname($module))) };
        
        print "\tCannot create directory: $@" if($@);
                                                    
        copy($module, catdir($master, dirname($module) ));
    }
                              
    foreach my $module (@mod)
    {
        print "\tModule ",basename($module);
        
        ($check{$module} == 1) ? print " successfully added to ramfs.\n" : 
                                 print " could not be found for kernel '$kernel'.\n";
    }
    return 1;
}

sub add_executables
{
    my $master = shift;
    my @binaries = @_;
    my @libraries = ();
    
    foreach my $prog (@binaries)
    {
        $prog = `which $prog` unless $prog =~ /\//;
        chomp($prog);
        if(! -x $prog)
        {
            print "\tFile '$prog' is not executable.\n";
            next;
        }
        push @libraries, `ldd $prog`;
    }
    
    @libraries = map { @{[split(/=>/,$_)]}[1] } @libraries;
    
    foreach my $entry (@binaries)
    {
        eval { mkpath(catdir($master,dirname($entry))) };
        if ($@) 
        {
            print "\tCannot create directory: $@";
        }
        unlink(catfile($master,$entry)); #remove existing entry before 
        if(copy($entry, catdir($master, dirname($entry) )) != 0)
        {
            print "\tFile '$entry' successfully added to ramfs.\n";
        }
        else
        {
            print "\tFile '$entry' cannot be copied to ramfs: $!\n";
        }
    }
    return @libraries ? add_libraries(@libraries) : 1;
}

sub add_libraries
{
    my $master = shift;
    my @libraries = @_;
    my %done;
    
    foreach my $lib (@libraries)
    {
        chomp($lib);
        next if($lib =~ /^$/);
        next unless $lib =~ /\//;
        $lib =~ s/^\s*([^\s]*).*/$1/;
        next if(defined $done{$lib});
        $done{$lib} = 1;
        
        if(! -e $lib)
        {
            print "\tLibrary '$lib' does not exist.\n";
            next;
        }
        
        eval { mkpath(catdir($master,dirname($lib))) };
        if ($@) 
        {
            print "\tCannot create directory: $@";
        }
        if(copy($lib, catdir($master, dirname($lib) )) != 0)
        {
            print "\tLibrary '$lib' successfully added to ramfs.\n";
        }
        else
        {
            print "\tLibrary '$lib' cannot be copied to ramfs: $!\n";
        }
        
        if(-x $lib)
        {
            my @temp = `ldd $lib`;
            @temp = map { @{[split(/=>/,$_)]}[1] } @temp;
            map { chomp } @temp;
            map { s/^\s*(\/[^\s]*)?.*/$1/ } @temp;
            push @libraries, @temp;
        }
    }
    return 1;
}

__DATA__

MASTER_RAMFS  /boot/initrd
KBOOT_RAMFS   /boot/kboot-root-glibc.cpio.gz
IMAGE_RAMFS   /boot/initrd-kboot.cpio.gz
TEMP_DIR      /tmp/KBOOT

############### the module dependent part ##################
KERNEL        /boot/vmlinuz
MODULE        e100
MODULE        reiserfs

##############  executables and libraries ##################
EXEC          uname
EXEC          losetup

LIB           /lib/libpam_misc.so.0
