/*!
 * (C) Copyright IBM Corp. 2006, 2006
 *
 * \file ui_ssh.c
 * \brief kboot user interface module for remote ssh-clients
 *
 * $Id: ui_ssh.c,v 1.10 2006/06/27 12:55:46 loehr Exp $
 *
 * Author(s): Swen Schillig   (swen@vnet.ibm.com)
 *
 */

#include <unistd.h>        //->unlink
#include <string.h>        //->strlen
#include <libgen.h>        //->basename
#include <syslog.h>        //->syslog
#include "ssh.h"
#include "config.h"
#include "kboot.h"

#define DROPBEAR_PIDFILE "/var/run/dropbear.pid"

char *arg0; //!< global variable with pointer to argv[0]

/*
 * get value for key from a string list
 *
 * \param[in] key  :to search a value for
 *            list :of strings where to search for key and value
 * \return    CFG_RETURN_OK if the user entered the correct password,
 *            CFG_RETURN_ERROR otherwise
 */
char *get_value(char const *key, char **list)
{
    char *t_ptr = NULL;

    if(list == NULL) return(NULL);

    while((t_ptr = *list++) != NULL)
    {
        if(strncmp(key,t_ptr,strlen(key)) != 0)
            continue;
        t_ptr += strlen(key);

        if(*t_ptr == '=')           //proper assignment
            t_ptr++;
        else if(*t_ptr == '\0')     //blank instead of '=' separator
            t_ptr = *list++;

        return(t_ptr);
    }
    return(NULL);
}


/*
 * assemble cmd-line (host-key file) parameter for SSH server
 *
 * \param[in]  type     : DSS or RSA
 *             key_param: file name string from cmd-line
 * \param[out] key_param: final cmd-line parameter for SSH server
 *
 * \return     ON SUCCESS final cmd-line parameter for SSH server
 *             ELSE       NULL
 */
char * assemble_key_param(char const *type, char *key_param, char *portnumber)
{
    char *temp    = NULL; //temporary string pointer for file names
    char *res_str = NULL; //resulting key_param string
    char *info    = NULL; //dummy pointer for comp_load function
    char *error   = NULL; //dummy pointer for comp_load function
    char *ex_str  = NULL; //execution string for 'system()'
    int  ex_code;        //execution return code
    char *def_str = (strcmp(type,"DSS") == 0) ? 
	    SSH_DSS_HOST_KEY : SSH_RSA_HOST_KEY; //default host key location


    cfg_strinit(&res_str);
    cfg_strinit(&temp);
    cfg_strinit(&ex_str);

    cfg_strprintf(&temp,"%s.%s", def_str, portnumber);

    if(comp_load(temp, key_param, &info, &error) == 0)
    {
	//temporary file name for key conversion
        cfg_strprintf(&res_str, "%s.tmp", temp);  

        cfg_strprintf(&ex_str,"%s openssh dropbear %s %s 2>/dev/null", 
	    SSH_KEY_CONVERT, temp, res_str);

	//convert openssh key to dropbear format
        ex_code = system(ex_str);            

        if(WEXITSTATUS(ex_code) > 0)         //error for key conversion
        {
            cfg_strprintf(&ex_str,"%s dropbear dropbear %s %s 2>/dev/null", 
		SSH_KEY_CONVERT, temp, res_str);

	    //check if key is a dropbear key already
            ex_code = system(ex_str);        
        }

	//conversion and renaming ok
        if( (WEXITSTATUS(ex_code) == 0) && (rename(res_str, temp) == 0) )  
        {
            cfg_strprintf(&res_str,"%s %s", 
		(strcmp(type,"DSS") == 0) ? "-d" : "-r", temp);

	    //free memory for process specific key file name
            cfg_strfree(&temp);                      
	    cfg_strfree(&ex_str);    
            return(res_str);
        }
    }

    //something went wrong (conversion || renaming || execution || loading)
    cfg_strprintf(&res_str,"%s %s", 
	(strcmp(type,"DSS") == 0) ? "-d" : "-r", def_str);

    cfg_strfree(&temp);
    cfg_strfree(&ex_str);    
    return(res_str);
}


int main(int argc,char **argv)
{
    char const *cmd_name = basename(argv[0]);
    char *portnumber     = get_value("port",argv) ;
    char *rsa_key        = get_value("rsa_key",argv);
    char *dss_key        = get_value("dss_key",argv);

    char *temp_fn = NULL;             //pid-file name
    char *ex_str  = NULL;             //execution string

    FILE *temp_fh = NULL;             //pid-file handle
    
    struct stat buf;                  //result buffer for stat() call

    pid_t ssh_pid;                    //sshd-pid
    int   ex_code;                    //execution return code

    openlog(cmd_name, LOG_PID, LOG_USER);

    //assign port number
    portnumber   = portnumber   ? portnumber : SSH_PORT;             

    syslog(LOG_INFO,"Starting SSH on port %s", portnumber);

    cfg_strinit(&temp_fn);
    cfg_strprintf(&temp_fn,"%s.%s",SSH_PID,portnumber);

    //check for existence of PID file
    if(stat(temp_fn,&buf) == 0)                                      
    {
        if((temp_fh = fopen(temp_fn,"r")))
        {
		//read PID for SSHD which is serving our port
		fscanf(temp_fh,"%ld",(long *) &ssh_pid);                 
		fclose(temp_fh);
		syslog(LOG_WARNING,"PID file exists already");
		//PID is running
		if(kill(ssh_pid, 0) == 0)                                 
		{
			syslog(LOG_ERR,"already running on port %s", 
			    portnumber);
			exit(10);
		}
		syslog(LOG_INFO,
		    "not running on port %s. Removing stale PID file\n", 
		    portnumber);
        }
	//delete stale PID file
        unlink(temp_fn);                                             
    }

    cfg_strfree(&temp_fn);

    if(! rsa_key)
    {
        rsa_key = malloc(strlen(SSH_RSA_HOST_KEY) + 5);
        sprintf(rsa_key,"-r %s", SSH_RSA_HOST_KEY);
    }
    else if(! (rsa_key = assemble_key_param("RSA", rsa_key, portnumber)))
        syslog(LOG_ERR,
	    "Cannot retrieve RSA host key file. Using default instead.\n");

    if(! dss_key)
    {
        dss_key = malloc(strlen(SSH_DSS_HOST_KEY) + 5);
        sprintf(dss_key,"-d %s", SSH_DSS_HOST_KEY);
    }
    else if(! (dss_key = assemble_key_param("DSS", dss_key, portnumber)))
        syslog(LOG_ERR,
	    "Cannot retrieve DSS host key file. Using default instead.\n");

    cfg_strinit(&ex_str);
    cfg_strprintf(&ex_str,"%s -a -p %s %s %s", 
	SSH_CMD, portnumber, dss_key, rsa_key);
    ex_code = system(ex_str);

    /* workaround for missing -P option */
    cfg_strprintf(&ex_str,"cp %s %s.%s", 
	DROPBEAR_PIDFILE, SSH_PID, portnumber);
    cfg_system(ex_str);

    cfg_strfree(&ex_str);

    closelog();
    return( (ex_code <= 0) ? ex_code : WEXITSTATUS(ex_code) );
}
